@@ -7,23 +7,153 @@ The tool can be used in two ways:
 """
 import argparse
 import ast
+import atexit
+import os
 import random
+import signal
 import subprocess
-import seaborn as sns
-import colored
+import sys
 import time
 from collections import defaultdict
 from datetime import datetime
 from pathlib import Path
+from typing import Dict, List
+import functools
 
+import colored
 import numpy as np
+import seaborn as sns
 
-from slurm_gpustat.daemon import Daemon
+from zsvision.zs_beartype import beartype
 
 # SLURM states which indicate that the node is not available for submitting jobs
 INACCESSIBLE = {"drain*", "down*", "drng", "drain", "down"}
 
 
+class Daemon:
+    """A Generic linux daemon base class for python 3.x.
+
+    This code is a Python3 port of Sander Marechal's Daemon module:
+    http://web.archive.org/web/20131017130434/http://www.jejik.com/articles/
+    2007/02/a_simple_unix_linux_daemon_in_python/
+
+    It's a little difficult to credit the author of the Python3 port, since the code was
+    published anonymously. The original can be found here:
+    http://web.archive.org/web/20131101191715/http://www.jejik.com/files/examples/
+    daemon3x.py
+    """
+
+    def __init__(self, pidfile):
+        self.pidfile = pidfile
+
+    def daemonize(self):
+        """Deamonize class. UNIX double fork mechanism."""
+        try:
+            pid = os.fork()
+            if pid > 0:
+                # exit first parent
+                sys.exit(0)
+        except OSError as err:
+            sys.stderr.write('fork #1 failed: {0}\n'.format(err))
+            sys.exit(1)
+
+        # decouple from parent environment
+        os.chdir('/')
+        os.setsid()
+        os.umask(0)
+
+        # do second fork
+        try:
+            pid = os.fork()
+            if pid > 0:
+                # exit from second parent
+                sys.exit(0)
+        except OSError as err:
+            sys.stderr.write('fork #2 failed: {0}\n'.format(err))
+            sys.exit(1)
+
+        # redirect standard file descriptors
+        sys.stdout.flush()
+        sys.stderr.flush()
+        si = open(os.devnull, 'r')
+        so = open(os.devnull, 'a+')
+        se = open(os.devnull, 'a+')
+
+        os.dup2(si.fileno(), sys.stdin.fileno())
+        os.dup2(so.fileno(), sys.stdout.fileno())
+        os.dup2(se.fileno(), sys.stderr.fileno())
+
+        # write pidfile
+        atexit.register(self.delpid)
+
+        pid = str(os.getpid())
+        with open(self.pidfile, 'w+') as f:
+            f.write(pid + '\n')
+
+    def delpid(self):
+        os.remove(self.pidfile)
+
+    def start(self):
+        """Start the daemon."""
+
+        # Check for a pidfile to see if the daemon already runs
+        try:
+            with open(self.pidfile, 'r') as pf:
+
+                pid = int(pf.read().strip())
+        except IOError:
+            pid = None
+
+        if pid:
+            message = "pidfile {0} already exists. Is the daemon already running?\n"
+            sys.stderr.write(message.format(self.pidfile))
+            sys.exit(1)
+
+        # Start the daemon
+        self.daemonize()
+        self.run()
+
+    def stop(self):
+        """Stop the daemon."""
+
+        # Get the pid from the pidfile
+        try:
+            with open(self.pidfile, 'r') as pf:
+                pid = int(pf.read().strip())
+        except IOError:
+            pid = None
+
+        if not pid:
+            message = "pidfile {0} does not exist. Daemon not running?\n"
+            sys.stderr.write(message.format(self.pidfile))
+            return  # not an error in a restart
+
+        # Try killing the daemon process
+        try:
+            while 1:
+                os.kill(pid, signal.SIGTERM)
+                time.sleep(0.1)
+        except OSError as err:
+            e = str(err.args)
+            if e.find("No such process") > 0:
+                if os.path.exists(self.pidfile):
+                    os.remove(self.pidfile)
+            else:
+                print(str(err.args))
+                sys.exit(1)
+
+    def restart(self):
+        """Restart the daemon."""
+        self.stop()
+        self.start()
+
+    def run(self):
+        """You should override this method when you subclass Daemon.
+        It will be called after the process has been daemonized by
+        start() or restart()."""
+        raise NotImplementedError("Must override this class")
+
+
 class GPUStatDaemon(Daemon):
     """A lightweight daemon which intermittently logs gpu usage to a text file.
     """
@@ -106,16 +236,18 @@ def historical_summary(data):
         data (list): the data structure deserialized from the daemon log file (this is
             the output of the GPUStatDaemon.deserialize_usage() function.)
     """
-    first_ts, last_ts = data[0]["timestamp"], data[1]["timestamp"]
+    first_ts, last_ts = data[0]["timestamp"], data[-1]["timestamp"]
     print(f"Historical data contains {len(data)} samples ({first_ts} to {last_ts})")
     latest_usage = data[-1]["usage"]
     users, gpu_types = set(), set()
     for user, resources in latest_usage.items():
         users.add(user)
         gpu_types.update(set(resources.keys()))
-    history = {user: {gpu_type: [] for gpu_type in gpu_types} for user in users}
+    history = {}
     for row in data:
         for user, subdict in row["usage"].items():
+            if user not in history:
+                history[user] = {gpu_type: [] for gpu_type in gpu_types}
             type_counts = {key: sum(val.values()) for key, val in subdict.items()}
             for gpu_type in gpu_types:
                 history[user][gpu_type].append(type_counts.get(gpu_type, 0))
@@ -211,11 +343,12 @@ def parse_cmd(cmd):
     return [x for x in output.split("\n") if x]
 
 
-def node_states():
+@beartype
+def node_states() -> Dict[str, str]:
     """Query SLURM for the state of each managed node.
 
     Returns:
-        (dict[str: str]): a mapping between node names and SLURM states.
+        a mapping between node names and SLURM states.
     """
     cmd = "sinfo --noheader"
     rows = parse_cmd(cmd)
@@ -228,16 +361,44 @@ def node_states():
     return states
 
 
-def parse_all_gpus(default_gpus=4):
+@beartype
+@functools.lru_cache(maxsize=64, typed=True)
+def occupancy_stats_for_node(node: str) -> Dict[str, int]:
+    """Query SLURM for the occupancy of a given node.
+
+    Args:
+        (node): the name of the node to query
+
+    Returns:
+        a mapping between node names and occupancy stats.
+    """
+    cmd = f"scontrol show node {node}"
+    rows = [x.strip() for x in parse_cmd(cmd)]
+    keys = ("AllocTRES", "CfgTRES")
+    metrics = {}
+    for row in rows:
+        for key in keys:
+            if row.startswith(key):
+                row = row.replace(f"{key}=", "")
+                tokens = row.split(",")
+                metrics[key] = {x.split("=")[0]: x.split("=")[1] for x in tokens}
+    occupancy = {}
+    for metric, alloc_val in metrics["AllocTRES"].items():
+        cfg_val = metrics["CfgTRES"][metric]
+        occupancy[metric] = f"{alloc_val}/{cfg_val}"
+    return occupancy
+
+
+@beartype
+def parse_all_gpus(default_gpus: int = 4) -> Dict[str, List]:
     """Query SLURM for the number and types of GPUs under management.
 
     Args:
-        default_gpus (int :: 4): The number of GPUs estimated for nodes that have
-            incomplete SLURM meta data.
+        default_gpus: The number of GPUs estimated for nodes that have incomplete SLURM
+            meta data.
 
     Returns:
-        (dict[str:list]): a mapping between node names and a list of the GPUs that they
-            have available.
+        a mapping between node names and a list of the GPUs that they have available.
     """
     cmd = "sinfo -o '%50N|%30G' --noheader"
     rows = parse_cmd(cmd)
@@ -258,23 +419,25 @@ def parse_all_gpus(default_gpus=4):
     return resources
 
 
-def resource_by_type(resources):
+@beartype
+def resource_by_type(resources: Dict[str, Dict]) -> Dict[str, Dict]:
     """Determine the cluster capacity by gpu type
 
     Args:
-        resources (dict): a summary of the cluster resources, organised by node name.
+        resources: a summary of the cluster resources, organised by node name.
 
-    Args:
-        resources (dict): a summary of the cluster resources, organised by gpu type
+    Returns:
+        resources: a summary of the cluster resources, organised by gpu type
     """
-    by_type = defaultdict(lambda: 0)
-    for specs in resources.values():
+    by_type = defaultdict(list)
+    for node, specs in resources.items():
         for spec in specs:
-            by_type[spec["type"]] += spec["count"]
+            by_type[spec["type"]].append({"node": node, "count": spec["count"]})
     return by_type
 
 
-def summary_by_type(resources, tag):
+@beartype
+def summary_by_type(resources: Dict[str, Dict], tag: str):
     """Print out out a summary of cluster resources, organised by gpu type.
 
     Args:
@@ -282,15 +445,17 @@ def summary_by_type(resources, tag):
         tag (str): a term that will be included in the printed summary.
     """
     by_type = resource_by_type(resources)
-    total = sum(by_type.values())
+    total = sum(x["count"] for sublist in by_type.values() for x in sublist)
     agg_str = []
-    for key, val in sorted(by_type.items(), key=lambda x: x[1]):
-        agg_str.append(f"{val} {key} gpus")
+    for key, val in sorted(by_type.items(), key=lambda x: sum(y["count"] for y in x[1])):
+        gpu_count = sum(x["count"] for x in val)
+        agg_str.append(f"{gpu_count} {key} gpus")
     print(f"There are a total of {total} gpus [{tag}]")
     print("\n".join(agg_str))
 
 
-def summary(mode, resources=None, states=None):
+@beartype
+def summary(mode: str, resources: Dict[str, Dict] = None, states: Dict[str, List] = None):
     """Generate a printed summary of the cluster resources.
 
     Args:
@@ -312,7 +477,8 @@ def summary(mode, resources=None, states=None):
     summary_by_type(res, tag=mode)
 
 
-def gpu_usage(resources):
+@beartype
+def gpu_usage(resources: Dict[str, Dict]) -> Dict[str, Dict]:
     """Build a data structure of the cluster resource usage, organised by user.
 
     Args:
@@ -338,6 +504,10 @@ def gpu_usage(resources):
             gpu_type = gpu_count_tokens[1]
         node_names = parse_node_names(node_str)
         for node_name in node_names:
+            # If a node still has jobs running but is draining, it will not be present
+            # in the "available" resources, so we ignore it
+            if node_name not in resources:
+                continue
             node_gpu_types = [x["type"] for x in resources[node_name]]
             if gpu_type is None:
                 if len(node_gpu_types) != 1:
@@ -355,11 +525,12 @@ def gpu_usage(resources):
     return usage
 
 
-def in_use(resources=None):
+@beartype
+def in_use(resources: Dict[str, Dict] = None):
     """Print a short summary of the resources that are currently used by each user.
 
     Args:
-        resources (dict :: None): a summary of cluster resources, organised by node name.
+        resources: a summary of cluster resources, organised by node name.
     """
     if not resources:
         resources = parse_all_gpus()
@@ -374,13 +545,18 @@ def in_use(resources=None):
         print(f"{user:10s} [{total}] {summary_str}")
 
 
-def available(resources=None, states=None):
+@beartype
+def available(
+        resources: Dict[str, Dict] = None,
+        states: Dict[str, str] = None,
+        verbose: bool = False,
+):
     """Print a short summary of resources available on the cluster.
 
     Args:
-        resources (dict :: None): a summary of cluster resources, organised by node name.
-        states (dict[str: str] :: None): a mapping between node names and SLURM states.
-    
+        resources: a summary of cluster resources, organised by node name.
+        states: a mapping between node names and SLURM states.
+
     NOTES: Some systems allow users to share GPUs.  The logic below amounts to a
     conservative estimate of how many GPUs are available.  The algorithm is:
 
@@ -403,22 +579,38 @@ def available(resources=None, states=None):
                 count = max(count - user_gpu_count, 0)
                 res[node_name][resource_idx]["count"] = count
     by_type = resource_by_type(res)
-    total = sum(by_type.values())
+    total = sum(x["count"] for sublist in by_type.values() for x in sublist)
     print(f"There are {total} gpus available:")
-    for key, val in by_type.items():
-        print(f"{key}: {val}")
+    for key, counts_for_gpu_type in by_type.items():
+        gpu_count = sum(x["count"] for x in counts_for_gpu_type)
+        tail = ""
+        if verbose:
+            summary = []
+            for x in counts_for_gpu_type:
+                node, count = x["node"], x["count"]
+                if count:
+                    occupancy = occupancy_stats_for_node(node)
+                    users = [user for user in usage if node in usage[user].get(key, [])]
+                    details = [f"{key}: {val}" for key, val in sorted(occupancy.items())]
+                    details = f"[{', '.join(details)}] [{','.join(users)}]"
+                    summary.append(f"\n -> {node}: {count} {key} {details}")
+            tail = " ".join(summary)
+        print(f"{key}: {gpu_count} available {tail}")
 
 
-def all_info():
+@beartype
+def all_info(color: int, verbose: bool):
     """Print a collection of summaries about SLURM gpu usage, including: all nodes
     managed by the cluster, nodes that are currently accesible and gpu usage for each
     active user.
     """
-    colors = sns.color_palette("hls", 8).as_hex()
-    divider = colored.stylize("---------------------------------", colored.fg(colors[7]))
+    divider, slurm_str = "---------------------------------", "SLURM"
+    if color:
+        colors = sns.color_palette("hls", 8).as_hex()
+        divider = colored.stylize(divider, colored.fg(colors[7]))
+        slurm_str = colored.stylize(slurm_str, colored.fg(colors[0]))
     print(divider)
-    slurm = colored.stylize("SLURM", colored.fg(colors[0]))
-    print(f"Under {slurm} management")
+    print(f"Under {slurm_str} management")
     print(divider)
     resources = parse_all_gpus()
     states = node_states()
@@ -427,7 +619,7 @@ def all_info():
         print(divider)
     in_use(resources)
     print(divider)
-    available(resources=resources, states=states)
+    available(resources=resources, states=states, verbose=verbose)
     print(divider)
 
 
@@ -448,18 +640,21 @@ def main():
                         help="the location where the daemon PID file will be stored")
     parser.add_argument("--daemon_log_interval", type=int, default=43200,
                         help="time interval (secs) between stat logging (default 12 hrs)")
+    parser.add_argument("--color", type=int, default=1, help="color output")
+    parser.add_argument("--verbose", action="store_true",
+                        help="provide a more detailed breakdown of resources")
     args = parser.parse_args()
 
     if args.action == "current":
-        all_info()
+        all_info(color=args.color, verbose=args.verbose)
     elif args.action == "history":
         data = GPUStatDaemon.deserialize_usage(args.log_path)
         historical_summary(data)
     elif args.action.startswith("daemon"):
         daemon = GPUStatDaemon(
             log_path=args.log_path,
-            log_interval=args.daemon_log_interval,
             pidfile=args.gpustat_pid,
+            log_interval=args.daemon_log_interval,
         )
         if args.action == "daemon-start":
             print("Starting daemon")
